"""
Command line interface
"""

from os.path import split, exists, join
from os import mkdir, remove
import argparse
import gzip
import shutil
from future.standard_library import install_aliases
install_aliases()
from urllib.request import urlopen
from urllib.error import HTTPError

import pyensembl
import agfusion
from agfusion import exceptions
from agfusion.utils import AGFUSION_DB_URL, AVAILABLE_ENSEMBL_SPECIES, GENOME_SHORTCUTS


def list_available_databases():
    """
    List the available databases that can be downloaded.
    """

    print('\n')
    print(
        '{:<10}\t\t{:<5}\t\t{:<20}'
        .format('Species', 'Release', 'Shortcut(s)')
    )
    for species, releases in AVAILABLE_ENSEMBL_SPECIES.items():
        for release in releases:
            shortcut = []
            for genome, data in GENOME_SHORTCUTS.items():
                if species in data and release in data:
                    shortcut.append(genome)
            print(
                '{:<10}\t\t{:<5}\t\t{:<20}'
                .format(species, release, ','.join(shortcut))
            )
    exit()

def downloaddb(args):
    """
    Download the AGFusion database from github
    """

    if args.genome is not None:
        if args.genome not in GENOME_SHORTCUTS:
            print('Invalid genome shortcut! Use -a to see available shortcuts.')
            exit()
        else:
            species = GENOME_SHORTCUTS[args.genome][0]
            release = str(GENOME_SHORTCUTS[args.genome][1])
    else:
        if args.species is None or args.release is None:
            print("Specify --species and --release or --genome!")
            exit()
        species = args.species
        release = str(args.release)

    file_path = join(
        args.dir,
        'agfusion.' + species + '.' + release + '.db.gz')

    print("Downloading the AGFusion database to {}...".format(file_path))

    db_url = AGFUSION_DB_URL + species + '.' + release + '.db.gz'

    try:
        response = urlopen(db_url)
    except HTTPError:
        print("Was unable to downloade the file {}!".format(db_url))
        exit()

    fout = open(file_path, 'wb')
    fout.write(response.read())
    fout.close()

    with gzip.open(file_path, 'rb') as f_in, open(file_path.replace('.gz', ''), 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)

    remove(file_path)

def annotate(gene5prime, junction5prime, gene3prime, junction3prime,
             agfusion_db, pyensembl_data, args, outdir=None, colors=None,
             rename=None, scale=None, batch_out_dir=None):
    """
    Annotate the gene fusion
    """

    fusion = agfusion.Fusion(
        gene5prime=gene5prime,
        gene5primejunction=junction5prime,
        gene3prime=gene3prime,
        gene3primejunction=junction3prime,
        db=agfusion_db,
        pyensembl_data=pyensembl_data,
        protein_databases=args.protein_databases,
        noncanonical=args.noncanonical
    )

    if batch_out_dir is not None:

        outdir = join(
            batch_out_dir,
            fusion.gene5prime.gene.name + '-' +
            str(junction5prime) + '_' +
            fusion.gene3prime.gene.name + '-' +
            str(junction3prime)
        )

    fusion.save_transcript_cdna(
        out_dir=outdir,
        middlestar=args.middlestar
    )
    fusion.save_transcript_cds(
        out_dir=outdir,
        middlestar=args.middlestar
    )
    fusion.save_proteins(
        out_dir=outdir,
        middlestar=args.middlestar
    )

    fusion.save_images(
        out_dir=outdir,
        scale=scale,
        colors=colors,
        rename=rename,
        fontsize=args.fontsize,
        height=args.height,
        width=args.width,
        dpi=args.dpi,
        no_domain_labels=args.no_domain_labels,
        plot_WT=args.WT,
        exclude=args.exclude_domain
        )
    fusion.save_tables(out_dir=outdir)

def batch_mode(args, agfusion_db, pyensembl_data, rename, colors):
    """
    Batch mode for annotation fusions from output from a fusion-finding
    algorithm
    """

    if not exists(args.out):
        mkdir(args.out)
    else:
        agfusion_db.logger.warn(
            'Output directory {} already exists! Overwriting...'
            .format(args.out)
        )

    if args.algorithm in agfusion.parsers:
        for fusion in agfusion.parsers[args.algorithm](args.file,agfusion_db.logger):

            try:
                annotate(
                    gene5prime=fusion['gene5prime'],
                    junction5prime=fusion['gene5prime_junction'],
                    gene3prime=fusion['gene3prime'],
                    junction3prime=fusion['gene3prime_junction'],
                    agfusion_db=agfusion_db,
                    pyensembl_data=pyensembl_data,
                    args=args,
                    colors=colors,
                    rename=rename,
                    scale=None,
                    batch_out_dir=args.out
                )
            except exceptions.GeneIDException as e:
                agfusion_db.logger.error(e)
            except exceptions.JunctionException as e:
                agfusion_db.logger.error(e)
            except exceptions.TooManyGenesException as e:
                agfusion_db.logger.error(e)
    else:
        agfusion_db.logger.error(
            '\'{}\' is not an available option for -a! Choose one of the ' \
            'following: {}.' \
            .format(
                args.algorithm,
                ','.join(agfusion.parsers.keys())
            )
        )
        exit()

def builddb(args):
    """
    Build a AGFusion database
    """

    agfusion_db = agfusion.AGFusionDBBManager(
        args.dir,
        args.species,
        args.release,
        args.pfam,
        args.server
    )

    agfusion_db.logger.info('Fetching alternative gene names...')

    agfusion_db.fetch_gene_names()

    agfusion_db.logger.info('Fetching transcript tables...')

    agfusion_db.fetch_transcript_table()

    agfusion_db.fetch_refseq_table()

    agfusion_db.logger.info('Fetching protein annotation data...')

    agfusion_db.fetch_protein_annotation()

def add_common_flags(parser):
    """
    Add commaond line flags that are common to multiple sub parsers
    """

    parser.add_argument(
        '-db',
        '--database',
        type=str,
        required=True,
        help='Path to the AGFusion database (e.g. --db /path/to/agfusion.homo_sapiens.87.db)'
    )
    parser.add_argument(
        '-o',
        '--out',
        type=str,
        required=True,
        help='Directory to save results'
    )
    parser.add_argument(
        '-nc',
        '--noncanonical',
        action='store_true',
        required=False,
        default=False,
        help='(Optional) Include non-canonical gene transcripts ' +
        'in the analysis (default False).'
    )
    parser.add_argument(
        '--protein_databases',
        type=str,
        required=False,
        nargs='+',
        default=['pfam', 'tmhmm'],
        help='(Optional) Space-delimited list of one or more protein ' +
        'feature databases to include when visualizing proteins. ' +
        'Options are: pfam, smart, superfamily, tigrfam, pfscan (Prosite_profiles), ' +
        'tmhmm (i.e. transmembrane), seg (low_complexity regions), ncoils ' +
        '(coiled coil regions), prints, ' +
        'pirsf, and signalp (signal peptide regions) ' +
        '(default: --protein_databases pfam and tmhmm).'
    )
    parser.add_argument(
        '--recolor',
        type=str,
        required=False,
        default=None,
        action='append',
        help='(Optional) Re-color a domain. Provide the original name of ' +
        'the domain then your color (semi-colon delimited, all in ' +
        'quotes). Can specify --recolor multiples for each domain. ' +
        '(e.g. --color \"Pkinase_Tyr;blue\" --color \"I-set;#006600\").')
    parser.add_argument(
        '--rename',
        type=str,
        required=False,
        default=None,
        action='append',
        help='(Optional) Rename a domain. Provide the original name of ' +
        'the domain then your new name (semi-colon delimited, ' +
        'all in quotes). Can specify --rename multiples for each ' +
        'domain. (e.g. --rename \"Pkinase_Tyr;Kinase\").')
    parser.add_argument(
        '--exclude_domain',
        type=str,
        required=False,
        default=[],
        nargs='+',
        help='(Optional) Exclude a certain domain(s) from plotting ' +
        'by providing a space-separated list of domain names.')
    parser.add_argument(
        '--type',
        type=str,
        required=False,
        default='png',
        help='(Optional) Image file type (png, jpeg, pdf). Default: png')
    parser.add_argument(
        '-w',
        '--width',
        type=int,
        required=False,
        default=10,
        help='(Optional) Image width in inches (default 10).')
    parser.add_argument(
        '-ht',
        '--height',
        type=int,
        required=False,
        default=3,
        help='(Optional) Image file height in inches (default 3).')
    parser.add_argument(
        '--dpi',
        type=int,
        required=False,
        default=None,
        help='(Optional) Dots per inch.')
    parser.add_argument(
        '--fontsize',
        type=int,
        required=False,
        default=12,
        help='(Optional) Fontsize (default 12).')
    parser.add_argument(
        '--WT',
        action='store_true',
        required=False,
        help='(Optional) Include this to plot wild-type architechtures ' +
        'of the 5\' and 3\' genes')
    parser.add_argument(
        '-ms',
        '--middlestar',
        action='store_true',
        required=False,
        help='(Optional) Insert a * at the junction position for the ' +
        'cdna, cds, and protein sequences (default False).')
    parser.add_argument(
        '-ndl',
        '--no_domain_labels',
        action='store_true',
        required=False,
        help='(Optional) Do not label domains.')
    parser.add_argument(
        '--debug',
        default=False,
        action='store_true',
        help='(Optional) Enable debugging logging.'
    )

def main():
    """
    Main function for processing command line options
    """

    parser = argparse.ArgumentParser(
        description='Annotate Gene Fusion (AGFusion)'
    )
    subparsers = parser.add_subparsers(
        help='AGFusion programs.',
        dest="subparser_name")

    annotate_parser = subparsers.add_parser(
        'annotate',
        help='Annotate and visualize a single fusion.')
    annotate_parser.add_argument(
        '-g5',
        '--gene5prime',
        type=str,
        required=True,
        help='5\' gene partner'
    )
    annotate_parser.add_argument(
        '-g3',
        '--gene3prime',
        type=str,
        required=True,
        help='3\' gene partner'
    )
    annotate_parser.add_argument(
        '-j5',
        '--junction5prime',
        type=int,
        required=True,
        help='Genomic location of predicted fuins for the 5\' gene partner. ' +
        'The 1-based position that is the last nucleotide included in ' +
        'the fusion before the junction.'
    )
    annotate_parser.add_argument(
        '-j3',
        '--junction3prime',
        type=int,
        required=True,
        help='Genomic location of predicted fuins for the 3\' gene partner. ' +
        'The 1-based position that is the first nucleotide included in ' +
        'the fusion after the junction.'
    )
    add_common_flags(annotate_parser)
    annotate_parser.add_argument(
        '--scale',
        type=int,
        required=False,
        default=-1,
        help='(Optional) Set maximum width (in amino acids) of the ' +
        'figure to rescale the fusion (default: max length of ' +
        'fusion product)')

    # batch file parser

    batch_parser = subparsers.add_parser(
        'batch',
        help='Annotate fusions from an output file from a fusion ' +
        'finding algorithm.')
    batch_parser.add_argument(
        '-f',
        '--file',
        type=str,
        required=True,
        help='Output file from fusion-finding algorithm.'
    )
    batch_parser.add_argument(
        '-a',
        '--algorithm',
        type=str,
        required=True,
        help='The fusion-finding algorithm. Can be one of the following: ' +
        ', '.join(agfusion.parsers.keys()) + '.'
    )
    add_common_flags(batch_parser)

    # download database

    database_parser = subparsers.add_parser(
        'download',
        help='Download database for a reference genome.')
    database_parser.add_argument(
        '-d',
        '--dir',
        type=str,
        default='',
        help='(Optional) Directory to the database will be downloaded ' +
        'to (defaults to current working directory).')
    database_parser.add_argument(
        '-g',
        '--genome',
        type=str,
        default=None,
        help='Specify the genome shortcut (e.g. hg19). To see all' +
        'available shortcuts run \'agfusion download -a\'. Either ' +
        'specify this or --species and --release.')
    database_parser.add_argument(
        '-s',
        '--species',
        type=str,
        default=None,
        help='The species (e.g. homo_sapiens).')
    database_parser.add_argument(
        '-r',
        '--release',
        type=int,
        default=None,
        help='The ensembl release (e.g. 87).')
    database_parser.add_argument(
        '-a',
        '--available',
        action='store_true',
        required=False,
        help='List available species and ensembl releases.')

    # build database parser

    build_database_parser = subparsers.add_parser(
        'build',
        help='Build database for a reference genome.')
    build_database_parser.add_argument(
        '-d',
        '--dir',
        type=str,
        required=True,
        help='Directory to write database file to.'
    )
    build_database_parser.add_argument(
        '-s',
        '--species',
        type=str,
        required=True,
        help='The species (e.g. homo_sapiens).'
    )
    build_database_parser.add_argument(
        '-r',
        '--release',
        type=int,
        required=True,
        help='The ensembl release (e.g. 87).'
    )
    build_database_parser.add_argument(
        '--pfam',
        type=str,
        required=True,
        help='File containing PFAM ID mappings.'
    )
    build_database_parser.add_argument(
        '--server',
        type=str,
        required=False,
        default='ensembldb.ensembl.org',
        help='(optional) Ensembl server (default ensembldb.ensembl.org)'
    )

    # agfusion version number

    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=agfusion.__version__
    )
    args = parser.parse_args()

    if args.subparser_name == 'build':
        builddb(args)
        exit()
    elif args.subparser_name == 'download':
        if args.available:
            list_available_databases()
        else:
            downloaddb(args)
        exit()

    # single or batch mode

    if not exists(args.out):
        mkdir(args.out)

    # if user does not specify a sqlite database then use the one provided
    # by the package

    db_file = split(args.database)[1]
    species = db_file.split('.')[1]
    release = db_file.split('.')[2]

    assert species in AVAILABLE_ENSEMBL_SPECIES, 'unsupported species!'

    agfusion_db = agfusion.AGFusionDB(args.database, debug=args.debug)
    agfusion_db.build = species + '_' + str(release)

    # get the pyensembl data

    pyensembl_data = pyensembl.EnsemblRelease(release, species)

    try:
        pyensembl_data.db
    except ValueError:
        agfusion_db.logger.error(
            "Missing pyensembl data. Run pyensembl install --release " +
            "{} --species {}".format(release, species)
        )
        exit()

    colors = {}
    rename = {}

    if args.recolor is not None:
        for i in args.recolor:
            pair = i.split(';')

            assert len(pair) == 2, " did not properly specify --colors"

            if pair[0] in colors:
                agfusion_db.logger.warn(
                    "You specified colors for {} twice."
                    .format(pair[0])
                )

            colors[pair[0]] = pair[1]

    if args.rename is not None:
        for i in args.rename:
            pair = i.split(';')

            assert len(pair) == 2, " did not properly specify --rename"

            if pair[0] in rename:
                agfusion_db.logger.warn(
                    "WARNING - you rename {} twice."
                    .format(pair[0])
                )

            rename[pair[0]] = pair[1]

    if args.subparser_name == 'annotate':
        annotate(
            gene5prime=args.gene5prime,
            junction5prime=args.junction5prime,
            gene3prime=args.gene3prime,
            junction3prime=args.junction3prime,
            agfusion_db=agfusion_db,
            pyensembl_data=pyensembl_data,
            args=args,
            outdir=args.out,
            colors=colors,
            rename=rename,
            scale=args.scale
        )
    elif args.subparser_name == 'batch':
        batch_mode(args, agfusion_db, pyensembl_data, rename, colors)



import itertools
import os
import re
import sys

from agfusion import utils, exceptions, plot
import pandas
from Bio import Seq, SeqIO, SeqRecord, SeqUtils
from Bio.Alphabet import generic_dna, generic_protein

# matplotlib.rcParams['interactive'] = False

from agfusion.utils import STANDARD_CHROMOSOMES, MIN_DOMAIN_LENGTH

class _Gene():
    """
    Stores the necessary information to specify the architecture of either
    wild-type genes or fusion gene.
    """

    def __init__(self, genes=None, junction=0, pyensembl_data=None,
                 genome='', gene5prime=False, db=None, noncanonical=False):
        """
        genes : str or list
            Provide one gene (str) or list of genes (list). In the case of a
            list of genes, it will seach each gene until it finds a hit.

        junction : int

        pyensembl_data : None

        genome : str

        gene5prime : bool

        db : None

        noncanonical : bool
        """

        if type(genes)==str:
            genes = [genes]
        elif type(genes)!=list:
            db.logger.error(
                'parameter \'gene\' to _Gene is not string or list: {}'
                .format(genes)
            )
            exit()
        if type(junction)!=int:
            db.logger.error(
                'parameter \'junction\' to _Gene is not int: {}'
                .format(junction)
            )
            exit()

        self.gene_found = False
        self.provided_transcript = False # indicates user if user provided transcript

        self.domains = []
        self.transcripts = {} # stores transcripts and their DB keys
        self.gene = None
        self.pyensembl_data = pyensembl_data
        self.genome = genome
        self.gene5prime = gene5prime
        self.db = db
        self.noncanonical = noncanonical

        # find the appropriate Ensembl gene ID

        for gene in genes:

            if re.findall('^NM_',gene) or re.findall('^NR_',gene):
                self._search_by_refseq(gene)
            if gene.isdigit() and not self.gene_found:
                self._search_as_entrez(gene)
            if re.findall('(^ENS.*G)', gene.upper()) and not self.gene_found:
                self._search_as_ensembl_id(gene)
            if re.findall('(^ENS.*T)', gene.upper()) and not self.gene_found:
                self._search_as_ensembl_transcript_id(gene)

            if not self.gene_found:

                # else check if it is a gene symbol

                self._search_by_symbol(gene)

                if not self.gene_found:
                    gene = gene.capitalize()
                    self._search_by_symbol(gene)

                if not self.gene_found:
                    gene = gene.upper()
                    self._search_by_symbol(gene)

            if self.gene_found:
                break

        # if gene has not been identified yet

        if not self.gene_found:
            raise exceptions.GeneIDException(','.join(genes))

        # else continue with processing

        self.junction = junction

        if not self.gene.contains(self.gene.contig,self.junction,self.junction):
            raise exceptions.JunctionException(self.gene.name, self.junction)

        # fetch the entrez gene id and canonical transcript id

        sqlite3_command = "SELECT * FROM " + self.db.build + " WHERE stable_id==\"" + self.gene.gene_id + "\""
        self.db.logger.debug('SQLite - ' + sqlite3_command)
        self.db.sqlite3_cursor.execute(
            sqlite3_command
        )
        tmp = self.db.sqlite3_cursor.fetchall()

        self.gene_id = tmp[0][0]
        self.entrez_id = tmp[0][2]
        self.canonical_transcript_id = tmp[0][4]

        # get the transcripts that will be processed and annotated

        if not noncanonical and not self.provided_transcript:

            # if only want the canonical and did not specify a certain transcript

            sqlite3_command = "SELECT * FROM " + self.db.build + "_transcript WHERE transcript_id==\"" + self.canonical_transcript_id + "\""
            self.db.logger.debug('SQLite - ' + sqlite3_command)
            self.db.sqlite3_cursor.execute(
                sqlite3_command
            )
            tmp = db.sqlite3_cursor.fetchall()
            self.transcripts[tmp[0][2]] = tmp[0][0]
        elif not noncanonical and self.provided_transcript:
            pass
        else:
            if self.provided_transcript and noncanonical:
                self.db.logger.warn("You provided a transcript ID as well as specified --noncanonical flag. Will process all the gene's transcripts.")

            # fetch transcript ids

            sqlite3_command = "SELECT * FROM " + self.db.build + "_transcript WHERE gene_id==\"" + self.gene_id + "\""
            self.db.logger.debug('SQLite - ' + sqlite3_command)
            self.db.sqlite3_cursor.execute(
                sqlite3_command
            )
            tmp = self.db.sqlite3_cursor.fetchall()

            for transcript in tmp:
                self.transcripts[transcript[2]] = transcript[0]

    def _search_as_ensembl_transcript_id(self,gene):
        # if it is ensembl transcript id

        if gene in self.pyensembl_data.transcript_ids():
            transcript = self.pyensembl_data.transcript_by_id(gene.upper())
            self.gene = transcript.gene

            self.db.logger.debug('Found Ensembl transcript entry for %s: %s' % (gene,self.gene.id))
            self.gene_found = True
            self.provided_transcript = True
        else:
            self.db.logger.debug('Found no Ensembl transcript entry for %s' % gene)

        sqlite3_command = "SELECT * FROM " + self.db.build + "_transcript WHERE transcript_stable_id==\"" + gene + "\""
        self.db.logger.debug('SQLite - ' + sqlite3_command)
        self.db.sqlite3_cursor.execute(
            sqlite3_command
        )
        tmp = self.db.sqlite3_cursor.fetchall()
        self.transcripts[tmp[0][2]] = tmp[0][0]

    def _search_as_ensembl_id(self,gene):
        # if it is ensembl gene id

        if gene in self.pyensembl_data.gene_ids():
            self.gene = self.pyensembl_data.gene_by_id(gene.upper())
            self.db.logger.debug('Found Enrez gene ID entry for %s: %s' % (gene,self.gene.id))
            self.gene_found = True
        else:
            self.db.logger.debug('Cannot find Ensembl gene id %s in database!' % gene)

    def _search_as_entrez(self,gene):
        # if it is an entrez gene ID

        sqlite3_command = "SELECT * FROM " + self.db.build + " WHERE entrez_id==\"" + gene + "\""
        self.db.logger.debug('SQLite - ' + sqlite3_command)
        self.db.sqlite3_cursor.execute(
            sqlite3_command
        )
        tmp = self.db.sqlite3_cursor.fetchall()

        if len(tmp)==1:
            self.gene = self.pyensembl_data.gene_by_id(tmp[0][1])
            self.db.logger.debug('Found Entrez gene ID entry for %s: %s' % (gene,self.gene.id))
            self.gene_found = True
        elif len(tmp)>1:

            self.db.logger.error('Found too many Entrez gene ID entries for %s!' % gene)
            for i in tmp:
                self.db.logger.error(i)
            sys.exit()
        else:
            self.db.logger.debug('Found no Entrez gene ID entry for %s' % gene)

    def _search_by_refseq(self,gene):


        # if it is RefSeq ID

        sqlite3_command = "SELECT * FROM " + self.db.build + "_refseq WHERE refseq_id==\"" + gene + "\""
        self.db.logger.debug('SQLite - ' + sqlite3_command)
        self.db.sqlite3_cursor.execute(
            sqlite3_command
        )
        tmp = self.db.sqlite3_cursor.fetchall()

        if len(tmp)==1:
            self.transcripts[tmp[0][1]] = tmp[0][0]
            self.db.logger.debug('Found RefSeq entry for %s' % gene)
            self.gene_found = True
            self.provided_transcript = True

            #fetch the gene

            transcript = self.pyensembl_data.transcript_by_id(tmp[0][1])
            self.gene = transcript.gene

        elif len(tmp)>1:
            self.db.logger.error('Found too many RefSeq entries for %s!' % gene)
            for i in tmp:
                self.db.logger.error(i)
            sys.exit()
        else:
            self.db.logger.debug('Found no RefSeq entry for %s' % gene)

    def _search_by_symbol(self,gene):
        if gene in self.pyensembl_data.gene_names():
            temp = self.pyensembl_data.genes_by_name(gene)

            if len(temp) > 1:

                # if too many ensembl gene IDs returned
                # use the one located on chromosomes

                for tmp_gene in temp:
                    if tmp_gene.contig in STANDARD_CHROMOSOMES:
                        self.gene = tmp_gene
                        self.gene_found = True

                if self.gene is None:
                    raise exceptions.TooManyGenesException(
                        gene,
                        [x.id for x in temp],
                        self.genome
                    )
            else:
                self.gene = temp[0]
                self.gene_found = True
                self.db.logger.debug('Found gene symbol entry for %s: %s' % (gene,self.gene.id))

class Fusion():
    """
    Generates the information needed for the gene fusion
    """

    def __init__(
            self,
            gene5prime=None, gene5primejunction=0,
            gene3prime=None, gene3primejunction=0,
            db=None, pyensembl_data=None, protein_databases=None,
            noncanonical=False,
            transcripts_5prime=None, transcripts_3prime=None):
        """
        gene5prime : str

        gene3prime : str

        gene5primejunction : int

        gene3primejunction : int

        db : None

        pyensembl_data : None

        protein_databases : None

        noncanonical : bool

        transcripts_5prime : str

        transcripts_3prime : str
        """

        self.db = db
        self.pyensembl_data = pyensembl_data

        # get the reference genom

        self.gene5prime = _Gene(
            genes=gene5prime,
            junction=gene5primejunction,
            pyensembl_data=pyensembl_data,
            gene5prime=True,
            db=db,
            noncanonical=noncanonical
        )

        self.gene3prime = _Gene(
            genes=gene3prime,
            junction=gene3primejunction,
            pyensembl_data=pyensembl_data,
            gene5prime=False,
            db=db,
            noncanonical=noncanonical
        )

        self.name = self.gene5prime.gene.name + '-' + self.gene3prime.gene.name

        # fetch which are canonical transcripts

        gene5prime_canonical = ''
        gene3prime_canonical = ''

        # construct all the fusion transcript combinations

        self.transcripts = {}

        for combo in list(itertools.product(
                        list(self.gene5prime.transcripts.keys()),
                        list(self.gene3prime.transcripts.keys()))):
            transcript1 = pyensembl_data.transcript_by_id(combo[0])
            transcript2 = pyensembl_data.transcript_by_id(combo[1])

            if transcripts_5prime is not None \
                    and transcript1.id not in transcripts_5prime:

                continue

            if transcripts_3prime is not None \
                    and transcript2.id not in transcripts_3prime:

                continue

            # skip if the junction is outside the range of either transcript

            name = transcript1.id + '-' + transcript2.id

            if not transcript1.contains(transcript1.contig,self.gene5prime.junction,self.gene5prime.junction):
                self.transcripts[name] = FusionTranscript(
                    transcript1=transcript1,
                    transcript2=transcript2,
                    gene5prime=self.gene5prime,
                    gene3prime=self.gene3prime,
                    db=db,
                    protein_databases=protein_databases,
                    predict_effect=False
                )
                self.transcripts[name].effect = 'Outside transcript boundry'
                self.transcripts[name].has_coding_potential = False

            elif not transcript2.contains(transcript2.contig,self.gene3prime.junction,self.gene3prime.junction):
                self.transcripts[name] = FusionTranscript(
                    transcript1=transcript1,
                    transcript2=transcript2,
                    gene5prime=self.gene5prime,
                    gene3prime=self.gene3prime,
                    db=db,
                    protein_databases=protein_databases,
                    predict_effect=False
                )
                self.transcripts[name].effect='Outside transcript boundry'
                self.transcripts[name].has_coding_potential=False

            else:
                self.transcripts[name] = FusionTranscript(
                    transcript1=transcript1,
                    transcript2=transcript2,
                    gene5prime=self.gene5prime,
                    gene3prime=self.gene3prime,
                    db=db,
                    protein_databases=protein_databases,
                )

    def output_to_html(
            self, fontsize=12, dpi=90, colors={}, rename={},
            width=8, height=2, scale=0, mpld3=None):
        """
        Write figures into html format for plotting in web tool rather
        than writing figures to file

        Needs the mpld3 package to be used
        """

        dict_of_plots = list()
        plot_key = dict()

        for name, transcript in list(self.transcripts.items()):

            if not transcript.has_coding_potential:
                continue

            fig = plt.figure(figsize=(width, height), dpi=dpi, frameon=False)

            self._draw(
                fig=fig,
                name=name,
                transcript=transcript,
                scale=scale,
                fontsize=fontsize,
                colors=[],
                rename=[]
            )

            single_chart = dict()
            single_chart['id'] = name
            single_chart['json'] = json.dumps(mpld3.fig_to_dict(fig))
            dict_of_plots.append(single_chart)

            plot_key[name] = transcript.name

            plt.close(fig)
            plt.clf()

        return dict_of_plots, plot_key

    def save_images(
            self, out_dir='', file_type='png', fontsize=12, dpi=100,
            colors={}, rename={}, width=8, height=2, scale=0,
            no_domain_labels=True, plot_WT=False,exclude=None):
        """
        Save images of all fusion isoforms and write to file
        Also plot the wild type proteins if desired
        """

        assert file_type in ['png', 'pdf'], 'provided wrong file type'

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        if plot_WT:
            gene5prime_WT = os.path.join(
                out_dir,
                self.gene5prime.gene.gene_name
            )
            gene3prime_WT = os.path.join(
                out_dir,
                self.gene3prime.gene.gene_name
            )
            if not os.path.exists(gene5prime_WT):
                os.mkdir(gene5prime_WT)

            if not os.path.exists(gene3prime_WT):
                os.mkdir(gene3prime_WT)

        for name, transcript in list(self.transcripts.items()):

            if not transcript.has_coding_potential:
                continue

            filename = os.path.join(out_dir, name + '.' + file_type)

            pplot = plot.PlotFusionProtein(
                filename=filename,
                width=width,
                height=height,
                dpi=dpi,
                scale=scale,
                fontsize=fontsize,
                colors=colors,
                rename=rename,
                no_domain_labels=no_domain_labels,
                transcript=transcript,
                exclude=exclude
            )
            pplot.draw()
            pplot.save()

            filename = os.path.join(
                out_dir,
                name + '.exon.' + file_type
            )

            pplot = plot.PlotFusionExons(
                transcript=transcript,
                filename=filename,
                width=width,
                height=height,
                dpi=dpi,
                scale=scale,
                fontsize=fontsize
            )
            pplot.draw()
            pplot.save()

            if plot_WT:

                # plot exons

                filename = os.path.join(
                    gene5prime_WT,
                    transcript.transcript1.id + '.exon.' + file_type
                )

                pplot = plot.PlotWTExons(
                    ensembl_transcript=transcript.transcript1,
                    filename=filename,
                    width=width,
                    height=height,
                    dpi=dpi,
                    scale=scale,
                    fontsize=fontsize
                )
                pplot.draw()
                pplot.save()

                filename = os.path.join(
                    gene3prime_WT,
                    transcript.transcript2.id + '.exon.' + file_type
                )

                pplot = plot.PlotWTExons(
                    ensembl_transcript=transcript.transcript2,
                    filename=filename,
                    width=width,
                    height=height,
                    dpi=dpi,
                    scale=scale,
                    fontsize=fontsize
                )
                pplot.draw()
                pplot.save()

                # plot proteins

                filename = os.path.join(
                    gene5prime_WT,
                    transcript.transcript1.id + '.' + file_type
                )

                pplot = plot.PlotWTProtein(
                    ensembl_transcript=transcript.transcript1,
                    filename=filename,
                    width=width,
                    height=height,
                    dpi=dpi,
                    scale=scale,
                    fontsize=fontsize,
                    colors=colors,
                    rename=rename,
                    no_domain_labels=no_domain_labels,
                    transcript=transcript,
                    exclude=exclude
                )
                pplot.draw()
                pplot.save()

                filename = os.path.join(
                    gene3prime_WT,
                    transcript.transcript2.id + '.' + file_type
                )

                pplot = plot.PlotWTProtein(
                    ensembl_transcript=transcript.transcript2,
                    filename=filename,
                    width=width,
                    height=height,
                    dpi=dpi,
                    scale=scale,
                    fontsize=fontsize,
                    colors=colors,
                    rename=rename,
                    no_domain_labels=no_domain_labels,
                    transcript=transcript,
                    exclude=exclude
                )
                pplot.draw()
                pplot.save()

    def save_transcript_cdna(self, out_dir='.', middlestar=False):
        """
        Save the cDNA sequences for all fusion isoforms to a fasta file
        """

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        fout = open(
            os.path.join(
                out_dir,
                self.name + '_cdna.fa'
            ),
            'w'
        )

        for name, transcript in list(self.transcripts.items()):

            if transcript.cdna is not None:

                if middlestar:
                    temp = str(transcript.cdna.seq)
                    temp = temp[:transcript.transcript_cdna_junction_5prime] + '*' + temp[transcript.transcript_cdna_junction_5prime:]
                    transcript.cdna.seq = Seq.Seq(temp,generic_dna)

                SeqIO.write(transcript.cdna,fout,"fasta")
            else:
                cdna = SeqRecord.SeqRecord(
                    Seq.Seq("",generic_dna),
                    id=transcript.name,
                    name=transcript.name,
                    description="No cDNA, fusion junction outside transcript(s) boundary"
                )
                SeqIO.write(cdna,fout,"fasta")

        fout.close()

    def save_transcript_cds(self, out_dir='.', middlestar=False):
        """
        Save the CDS sequences for all fusion isoforms to a fasta file
        """

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        #check if any transcripts have coding potential

        n=0
        for name, transcript in list(self.transcripts.items()):

            if transcript.cds is not None:
                n+=1

        if n == 0:
            self.db.logger.debug('The %s fusion does not produce any protein coding transcripts. No cds.fa file will be written' % self.name)
            return

        fout = open(
            os.path.join(
                out_dir,
                self.name + '_cds.fa'
            ),
            'w'
        )

        for name, transcript in list(self.transcripts.items()):

            if transcript.cds is not None:

                if middlestar:
                    temp = str(transcript.cds.seq)
                    temp = temp[:transcript.transcript_cds_junction_5prime] + '*' + temp[transcript.transcript_cds_junction_5prime:]
                    transcript.cds.seq = Seq.Seq(temp,generic_dna)

                SeqIO.write(transcript.cds,fout,"fasta")

        fout.close()

    def save_proteins(self, out_dir='.', middlestar=False):
        """
        Save the protein sequences for all fusion isoforms to a fasta file
        """

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        # check if any transcripts have coding potential

        n = 0
        for name, transcript in list(self.transcripts.items()):

            if transcript.cds is not None:
                n += 1

        if n == 0:
            self.db.logger.debug('The %s fusion does not produce any protein coding transcripts. No proteins.fa file will be written' % self.name)
            return

        fout = open(
            os.path.join(
                out_dir,
                self.name + '_protein.fa'
            ),
            'w'
        )

        for name, transcript in list(self.transcripts.items()):

            if transcript.cds is not None:

                if middlestar:
                    temp = str(transcript.protein.seq)
                    temp = temp[:transcript.transcript_protein_junction_5prime] + '*' + temp[transcript.transcript_protein_junction_5prime:]
                    transcript.protein.seq = Seq.Seq(temp,generic_protein)

                SeqIO.write(transcript.protein,fout,"fasta")

        fout.close()

    def save_tables(self,out_dir='.',annotation='pfam'):

        #write table of fusion isoforms

        fout = open(
            os.path.join(
                out_dir,
                self.name + '.fusion_transcripts.txt'
            ),
            'w'
        )
        fout.write(
            '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %
            (
                "5\'_gene",
                "3\'_gene",
                "5\'_transcript",
                "3\'_transcript",
                "5\'_transcript_biotype",
                "3\'_transcript_biotype",
                "Fusion_effect",
                "Protein_length",
                "Protein_weight_(kD)"
            )
        )
        for name, transcript in list(self.transcripts.items()):

            if transcript.protein_length is None:
                protein_length = "NA"
            else:
                protein_length = transcript.protein_length

            if transcript.molecular_weight is None:
                molecular_weight = "NA"
            else:
                molecular_weight = transcript.molecular_weight

            fout.write(
                '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %
                (
                    transcript.gene5prime.gene.gene_name,
                    transcript.gene3prime.gene.gene_name,
                    transcript.transcript1.id,
                    transcript.transcript2.id,
                    transcript.transcript1.biotype,
                    transcript.transcript2.biotype,
                    transcript.effect,
                    protein_length,
                    molecular_weight
                )
            )
        fout.close()

        # write table containing protein domains for each fusion isoform

        fout = open(
            os.path.join(
                out_dir,
                self.name + '.protein_domains.txt'
            ),
            'w'
        )
        fout.write(
            '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %
            (
                "5\'_gene",
                "3\'_gene",
                "5\'_transcript",
                "3\'_transcript",
                "Domain_ID",
                "Domain_name",
                "Domain_description",
                "Protein_start",
                "Protein_end"
            )
        )

        for name, transcript in list(self.transcripts.items()):
            for domain in transcript.domains['fusion']:
                fout.write(
                    '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %
                    (
                        transcript.gene5prime.gene.gene_name,
                        transcript.gene3prime.gene.gene_name,
                        transcript.transcript1.id,
                        transcript.transcript2.id,
                        domain[0],
                        domain[1],
                        domain[2],
                        domain[3],
                        domain[4]
                    )
                )
        fout.close()

        # write tables of exon structure

        fout = open(
            os.path.join(
                out_dir,
                self.name + '.exons.txt'
            ),
            'w'
        )
        fout.write(
            '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %
            (
                "5\'_gene",
                "3\'_gene",
                "5\'_transcript",
                "3\'_transcript",
                "exon_gene_source",
                "exon_number",
                "exon_chr",
                "exon_start",
                "exon_end"
            )
        )
        for name, transcript in list(self.transcripts.items()):
            for exon in transcript.gene5prime_exon_intervals:
                fout.write(
                    '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %
                    (
                        transcript.gene5prime.gene.gene_name,
                        transcript.gene3prime.gene.gene_name,
                        transcript.transcript1.id,
                        transcript.transcript2.id,
                        '\'5 gene',
                        exon[2],
                        transcript.transcript1.contig,
                        exon[0],
                        exon[1]
                    )
                )

            for exon in transcript.gene3prime_exon_intervals:
                fout.write(
                    '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %
                    (
                        transcript.gene5prime.gene.gene_name,
                        transcript.gene3prime.gene.gene_name,
                        transcript.transcript1.id,
                        transcript.transcript2.id,
                        '\'3 gene',
                        exon[2],
                        transcript.transcript2.contig,
                        exon[0],
                        exon[1]
                    )
                )

        fout.close()


class FusionTranscript(object):
    """
    Generates the information needed for the gene fusion transctips
    """

    def __init__(
            self,
            transcript1=None, transcript2=None,
            gene5prime=None, gene3prime=None,
            db=None, protein_databases=None,
            predict_effect=True):

        self.transcript1 = transcript1
        self.transcript2 = transcript2
        self.gene5prime = gene5prime
        self.gene3prime = gene3prime
        self.protein_databases = protein_databases
        self.db = db

        self.name = self.transcript1.id + '-' + self.transcript2.id
        self.gene_names = self.gene5prime.gene.name + '-' + self.gene3prime.gene.name

        self.cds = None
        self.transcript_cds_junction_5prime = None
        self.transcript_cds_junction_3prime = None
        self.gene5prime_cds_intervals = []
        self.gene3prime_cds_intervals = []
        self.cds_3prime = None
        self.cds_3prime = None

        self.cdna = None
        self.cdna_5prime = None
        self.cdna_3prime = None
        self.transcript_cdna_junction_5prime = None
        self.transcript_cdna_junction_3prime = None
        self.gene5prime_exon_intervals = []
        self.gene3prime_exon_intervals = []

        self.protein = None
        self.protein_length = None
        self.molecular_weight = None
        self.domains = {
            transcript1.id: [],
            transcript2.id: [],
            'fusion': []
        }

        self.transcript_protein_junction_5prime = None
        self.transcript_protein_junction_3prime = None

        self.effect_5prime = 'exon'
        self.effect_3prime = 'exon'
        self.effect = ''
        self.has_coding_potential = False

        if predict_effect:
            self.predict_effect()

    def _annotate(self):
        """
        Annotate the gene fusion's protein using the protein annotaiton
        from its two genes
        """

        fusion_domains = []
        gene5prime_domains = []
        gene3prime_domains = []

        tmp_domains = []

        # fetch the translation ids

        sqlite3_command = "SELECT * FROM " + self.db.build + "_transcript WHERE transcript_stable_id==\"" + self.transcript1.id + "\""
        self.db.logger.debug('SQLite - ' + sqlite3_command)
        self.db.sqlite3_cursor.execute(
            sqlite3_command
        )
        gene5prime_translation_id = self.db.sqlite3_cursor.fetchall()[0][3]

        sqlite3_command = "SELECT * FROM " + self.db.build + "_transcript WHERE transcript_stable_id==\"" + self.transcript2.id + "\""
        self.db.logger.debug('SQLite - ' + sqlite3_command)
        self.db.sqlite3_cursor.execute(
            sqlite3_command
        )
        gene3prime_translation_id = self.db.sqlite3_cursor.fetchall()[0][3]

        for protein_database in self.protein_databases:

            # fetch protein annotation

            sqlite3_command = "SELECT * FROM " + self.db.build + "_" + protein_database + " WHERE translation_id==\"" + gene5prime_translation_id + "\""
            self.db.logger.debug('SQLite - ' + sqlite3_command)
            self.db.sqlite3_cursor.execute(
                sqlite3_command
            )
            tmp_domains += [list(x) for x in self.db.sqlite3_cursor.fetchall()]

        for d in tmp_domains:

            pfeature_ID = d[2]
            pfeature_name = d[6]
            pfeature_description = d[5]
            pfeature_start = int(d[3])
            pfeature_end = int(d[4])

            gene5prime_domains.append([
                pfeature_ID,
                pfeature_name,
                pfeature_description,
                pfeature_start,
                pfeature_end
            ])

            try:
                if self.transcript_protein_junction_5prime < pfeature_start:
                    continue
                elif self.transcript_protein_junction_5prime >= pfeature_end:

                    fusion_domains.append([
                        pfeature_ID,
                        pfeature_name,
                        pfeature_description,
                        pfeature_start,
                        pfeature_end
                    ])

                elif (self.transcript_protein_junction_5prime - pfeature_start) >= MIN_DOMAIN_LENGTH:

                    pfeature_end = self.transcript_protein_junction_5prime

                    fusion_domains.append([
                        pfeature_ID,
                        pfeature_name,
                        pfeature_description,
                        pfeature_start,
                        pfeature_end
                    ])
            except:
                import pdb; pdb.set_trace()

        # only find the 3' gene partner's domains if the fusion is in-frame

        if self.effect != 'out-of-frame':

            tmp_domains = []

            for database in self.protein_databases:

                sqlite3_command = "SELECT * FROM " + self.db.build + "_" + database + " WHERE translation_id==\"" + gene3prime_translation_id + "\""
                self.db.logger.debug('SQLite - ' + sqlite3_command)
                self.db.sqlite3_cursor.execute(
                    sqlite3_command
                )
                tmp_domains += [list(x) for x in self.db.sqlite3_cursor.fetchall()]

            for d in tmp_domains:

                pfeature_ID = d[2]
                pfeature_name = d[6]
                pfeature_description = d[5]
                pfeature_start = int(d[3])
                pfeature_end = int(d[4])

                gene3prime_domains.append([
                    pfeature_ID,
                    pfeature_name,
                    pfeature_description,
                    pfeature_start,
                    pfeature_end
                ])

                if self.transcript_protein_junction_3prime > pfeature_end:
                    continue
                elif self.transcript_protein_junction_3prime <= pfeature_start:

                    pfeature_start = (pfeature_start-self.transcript_protein_junction_3prime) + self.transcript_protein_junction_5prime
                    pfeature_end = (pfeature_end-self.transcript_protein_junction_3prime) + self.transcript_protein_junction_5prime

                    fusion_domains.append([
                        pfeature_ID,
                        pfeature_name,
                        pfeature_description,
                        pfeature_start,
                        pfeature_end
                    ])

                else:

                    pfeature_start = self.transcript_protein_junction_5prime
                    pfeature_end = (pfeature_end-self.transcript_protein_junction_3prime) + self.transcript_protein_junction_5prime

                    fusion_domains.append([
                        pfeature_ID,
                        pfeature_name,
                        pfeature_description,
                        pfeature_start,
                        pfeature_end
                    ])

                if pfeature_end < pfeature_start:
                    import pdb; pdb.set_trace()

        self.domains['fusion'] = fusion_domains
        self.domains[self.transcript1.id] = gene5prime_domains
        self.domains[self.transcript2.id] = gene3prime_domains

    def _fetch_protein(self):
        """
        Predict the potential protain amino acid sequence
        """

        self.transcript_protein_junction_5prime = 0
        self.transcript_protein_junction_3prime = 0

        self.protein_names = self.transcript1.protein_id + '-' + self.transcript2.protein_id

        self.transcript_protein_junction_5prime = int(self.transcript_cds_junction_5prime/3.)

        if (len(self.cds_5prime)/3.).is_integer() and (len(self.cds_3prime)/3.).is_integer():
            self.effect='in-frame'
            self.transcript_protein_junction_3prime = int(self.transcript_cds_junction_3prime/3.)
        elif round((len(self.cds_5prime)/3. % 1) + (len(self.cds_3prime)/3. % 1),2) == 1.0:
            self.effect = 'in-frame (with mutation)'
            self.transcript_protein_junction_3prime = int(self.transcript_cds_junction_3prime/3.)
        else:
            self.effect = 'out-of-frame'

        #translate CDS into protein and remove everything after the stop codon

        #trim the CDS sequence if fusion is out-of-frame

        if self.effect == 'out-of-frame':
            seq = self.cds.seq[0:3*int(len(self.cds.seq)/3)]

            protein_seq = seq.translate()
            protein_seq = protein_seq[0:protein_seq.find('*')]
        else:
            protein_seq = self.cds.seq.translate()
            protein_seq = protein_seq[0:protein_seq.find('*')]

        if (len(self.cds.seq) % 3) !=0:
            self.db.logger.warn('Fusion isoform effect is not out-of-frame but CDS is not a multiple of 3!')

        # predict molecular weight

        self.molecular_weight = SeqUtils.molecular_weight(protein_seq)/1000.

        # convert to a sequence record

        self.protein_length = len(protein_seq)

        self.protein = SeqRecord.SeqRecord(
            protein_seq,
            id=self.protein_names,
            name=self.protein_names,
            description="length=" + str(self.protein_length) + \
                ", kD: " + str(self.molecular_weight) + \
                ", transcripts: " + str(self.name) + \
                ", genes: " + str(self.gene_names) + \
                ", effect: " + self.effect
        )

    def _fetch_transcript_cds(self):
        """
        Predict the potential nucleotide sequence
        """

        self.transcript_cds_junction_5prime = 0
        self.transcript_cds_junction_3prime = 0

        #5prime transcript

        if self.transcript1.strand=="+":
            for cds in self.transcript1.coding_sequence_position_ranges:
                if self.gene5prime.junction >= cds[1]:
                    self.transcript_cds_junction_5prime += (cds[1] - cds[0] + 1)
                elif self.gene5prime.junction <= cds[0]:
                    break
                else:
                    self.transcript_cds_junction_5prime += (self.gene5prime.junction - cds[0] + 1)
                    break
        else:
            for cds in self.transcript1.coding_sequence_position_ranges:
                if self.gene5prime.junction <= cds[0]:
                    self.transcript_cds_junction_5prime += (cds[1] - cds[0] + 1)
                elif self.gene5prime.junction >= cds[1]:
                    break
                else:
                    self.transcript_cds_junction_5prime += (cds[1] - self.gene5prime.junction + 1)

        self.cds_5prime = self.transcript1.coding_sequence[0:self.transcript_cds_junction_5prime]

        if self.transcript2.strand=="+":
            for cds in self.transcript2.coding_sequence_position_ranges:
                if self.gene3prime.junction >= cds[1]:
                    self.transcript_cds_junction_3prime += (cds[1] - cds[0] + 1)
                elif self.gene3prime.junction <= cds[0]:
                    break
                else:
                    self.transcript_cds_junction_3prime += (self.gene3prime.junction - cds[0])
        else:
            for cds in self.transcript2.coding_sequence_position_ranges:
                if self.gene3prime.junction <= cds[0]:
                    self.transcript_cds_junction_3prime += (cds[1] - cds[0] + 1)
                elif self.gene3prime.junction >= cds[1]:
                    break
                else:
                    self.transcript_cds_junction_3prime += (cds[1] - self.gene3prime.junction)

        self.cds_3prime = self.transcript2.coding_sequence[self.transcript_cds_junction_3prime::]

        #create a sequence record

        if self.cds_5prime is not None and self.cds_3prime is not None:
            seq = self.cds_5prime + self.cds_3prime

        self.cds = SeqRecord.SeqRecord(
            Seq.Seq(seq,generic_dna),
            id=self.name,
            name=self.name,
            description="length=" + str(len(self.cds_5prime + self.cds_3prime)) + \
                ", genes: " + \
                str(self.transcript1.gene.name) + '-' + \
                str(self.transcript2.gene.name)
        )

    def _fetch_transcript_cdna(self):
        """
        Predict the potential nucleotide sequence
        """

        self.transcript_cdna_junction_5prime = 0
        self.transcript_cdna_junction_3prime = 0

        # get the 5prime transcript sequence and determine if junction is
        # within intron

        n = 0

        n_max = len(self.transcript1.exon_intervals)
        exons = self.transcript1.exon_intervals

        if self.transcript1.strand == "+":

            exon_count = 0

            for exon in exons:

                exon_count += 1

                # is in intron?

                if n == 0 and self.gene5prime.junction < exon[0]:
                    self.effect_5prime = 'intron (before cds)'

                if n == n_max and self.gene5prime.junction > exon[1]:
                    self.effect_5prime = 'intron (after cds)'
                elif self.gene5prime.junction > exon[1] and self.gene5prime.junction < exons[n+1][0]:
                    self.effect_5prime = 'intron'

                n += 1

                # get sequence

                if self.gene5prime.junction >= exon[1]:
                    self.transcript_cdna_junction_5prime += (exon[1] - exon[0] + 1)
                    self.gene5prime_exon_intervals.append([
                        exon[0],
                        exon[1],
                        exon_count
                    ])
                elif self.gene5prime.junction <= exon[0]:
                    break
                else:
                    self.transcript_cdna_junction_5prime += (self.gene5prime.junction - exon[0] + 1)
                    self.gene5prime_exon_intervals.append([
                        exon[0],
                        self.gene5prime.junction,
                        exon_count
                    ])
                    break
        else:

            exon_count = 0

            for exon in exons:

                exon_count += 1

                # is in intron?

                if n == 0 and self.gene5prime.junction > exon[1]:
                    self.effect_5prime = 'intron (before cds)'

                if n == n_max and self.gene5prime.junction < exon[0]:
                    self.effect_5prime = 'intron (after cds)'
                elif self.gene5prime.junction < exon[0] and self.gene5prime.junction > exons[n+1][1]:
                    self.effect_5prime = 'intron'

                n += 1

                # get sequence

                if self.gene5prime.junction <= exon[0]:
                    self.transcript_cdna_junction_5prime += (exon[1] - exon[0] + 1)
                    self.gene5prime_exon_intervals.append([
                        exon[0],
                        exon[1],
                        exon_count
                    ])
                elif self.gene5prime.junction >= exon[1]:
                    break
                else:
                    self.transcript_cdna_junction_5prime += (exon[1] - self.gene5prime.junction + 1)
                    self.gene5prime_exon_intervals.append([
                        self.gene5prime.junction,
                        exon[1],
                        exon_count
                    ])

        try:
            self.cdna_5prime = self.transcript1.sequence[0:self.transcript_cdna_junction_5prime]
        except TypeError:
            self.db.logger.warn('No cDNA sequence available for %s! ' \
                'Will not print cDNA sequence for the %s fusion. ' \
                'You might be working with an outdated pyensembl. ' \
                'Update the package and rerun \'pyensembl install\'' %
                (
                    str(self.gene5prime.gene.name),
                    self.gene_names
                )
            )

        # get the 3prime transcript sequence and determine if junction is
        # within intron

        n = 0
        n_max = len(self.transcript2.exon_intervals)
        exons = self.transcript2.exon_intervals

        if self.transcript2.strand == "+":

            exon_count = 0

            # get the exons in the fusion

            for exon in exons:
                exon_count += 1
                if self.gene3prime.junction >= exon[1]:
                    continue
                elif self.gene3prime.junction <= exon[0]:
                    self.gene3prime_exon_intervals.append([
                        exon[0],
                        exon[1],
                        exon_count
                    ])
                else:
                    self.gene3prime_exon_intervals.append([
                        self.gene3prime.junction,
                        exon[1],
                        exon_count
                    ])

            for exon in exons:

                # is in intron?

                if n == 0 and self.gene3prime.junction < exon[0]:
                    self.effect_3prime = 'intron (before cds)'

                if n == n_max and self.gene3prime.junction > exon[1]:
                    self.effect_3prime = 'intron (after cds)'
                elif self.gene3prime.junction > exon[1] and self.gene3prime.junction < exons[n+1][0]:
                    self.effect_3prime = 'intron'

                n += 1

                # get sequence

                if self.gene3prime.junction >= exon[1]:
                    self.transcript_cdna_junction_3prime += (exon[1] - exon[0] + 1)
                elif self.gene3prime.junction <= exon[0]:
                    break
                else:
                    self.transcript_cdna_junction_3prime += (self.gene3prime.junction - exon[0])
        else:

            exon_count = 0

            # get the exons in the fusion

            for exon in exons:
                exon_count += 1
                if self.gene3prime.junction <= exon[0]:
                    continue
                elif self.gene3prime.junction >= exon[1]:
                    self.gene3prime_exon_intervals.append([
                        exon[0],
                        exon[1],
                        exon_count
                    ])
                else:
                    self.gene3prime_exon_intervals.append([
                        exon[0],
                        self.gene3prime.junction,
                        exon_count
                    ])

            for exon in exons:

                # is in intron?

                if n == 0 and self.gene3prime.junction > exon[1]:
                    self.effect_3prime = 'intron (before cds)'

                if n == n_max and self.gene3prime.junction < exon[0]:
                    self.effect_3prime = 'intron (after cds)'
                elif self.gene3prime.junction < exon[0] and self.gene3prime.junction > exons[n+1][1]:
                    self.effect_3prime = 'intron'

                n += 1

                # get sequence

                if self.gene3prime.junction <= exon[0]:
                    self.transcript_cdna_junction_3prime += (exon[1] - exon[0] + 1)
                elif self.gene3prime.junction >= exon[1]:
                    break
                else:
                    self.transcript_cdna_junction_3prime += (exon[1] - self.gene3prime.junction)

        if self.cdna_5prime is not None:
            try:
                self.cdna_3prime = self.transcript2.sequence[
                    self.transcript_cdna_junction_3prime::
                ]
            except TypeError:
                self.db.logger.warn('No cDNA sequence available for %s! ' \
                    'Will not print cDNA sequence for the %s fusion. ' \
                    'You might be working with an outdated pyensembl. ' \
                    'Update the package and rerun \'pyensembl install\'' %
                    (
                        str(self.gene3prime.gene.name),
                        self.gene_names
                    )
                )

        if self.cdna_5prime is not None and self.cdna_3prime is not None:
            seq = self.cdna_5prime + self.cdna_3prime
            seq_length = str(len(seq))
        else:
            seq = ''
            seq_length = 'NA'

        self.cdna = SeqRecord.SeqRecord(
            Seq.Seq(seq, generic_dna),
            id=self.name,
            name=self.name,
            description="length=" + seq_length
        )

        # find out if the junction on either gene is with in the 5' or 3' UTR,
        # or if it exactly at the beginning or end of the UTR

        # the 5' gene

        if self.effect_5prime.find('intron') == -1:

            if self.transcript1.complete and \
                    self.transcript_cdna_junction_5prime < len(self.transcript1.five_prime_utr_sequence):
                self.effect_5prime='5UTR'
            elif self.transcript1.complete and \
                    self.transcript_cdna_junction_5prime == len(self.transcript1.five_prime_utr_sequence):
                self.effect_5prime='5UTR (end)'

            if self.transcript1.complete and \
                    (len(self.transcript1)-self.transcript_cdna_junction_5prime) < len(self.transcript1.three_prime_utr_sequence):
                self.effect_5prime='3UTR'
            elif self.transcript1.complete and \
                    (len(self.transcript1)-self.transcript_cdna_junction_5prime) == len(self.transcript1.three_prime_utr_sequence):
                self.effect_5prime='3UTR (start)'

        #the 3' gene

        if self.effect_3prime.find('intron')==-1:

            if self.transcript2.complete and \
                    self.transcript_cdna_junction_3prime < len(self.transcript2.five_prime_utr_sequence):
                self.effect_3prime='5UTR'
            elif self.transcript2.complete and \
                    self.transcript_cdna_junction_3prime == len(self.transcript2.five_prime_utr_sequence):
                self.effect_3prime='5UTR (end)'

            if self.transcript2.complete and \
                    (len(self.transcript2)-self.transcript_cdna_junction_3prime) < len(self.transcript2.three_prime_utr_sequence):
                self.effect_3prime='3UTR'
            elif self.transcript2.complete and \
                    (len(self.transcript2)-self.transcript_cdna_junction_3prime) == len(self.transcript2.three_prime_utr_sequence):
                self.effect_3prime='3UTR (start)'

        self.effect = self.effect_5prime + '-' + self.effect_3prime

    # def _check_if_in_intron(self):

    def predict_effect(self):
        """
        For all gene isoform combinations predict the effect of the fusion
        (e.g. in-frame, out-of-frame, etc...). Then if it has coding potential
        then annotate with domains
        """

        # check if within exon or UTR

        self._fetch_transcript_cdna()

        # check if within CDS and if it occurs at the very beginning or end of CDS

        if self.transcript1.complete and (self.effect_5prime.find('UTR')==-1):

            if self.effect_5prime.find('intron') != -1:

                # if in intron, is it between CDS regions?

                n = 0
                n_max = len(self.transcript1.coding_sequence_position_ranges)-1

                for cds in self.transcript1.coding_sequence_position_ranges:

                    if self.transcript1.strand == "+":
                        if n == 0 and self.gene5prime.junction < cds[0]:
                            self.effect_5prime = 'intron (before cds)'
                            break
                        elif n == n_max and self.gene5prime.junction > cds[1]:
                            self.effect_5prime = 'intron (after cds)'
                            break
                    else:
                        if n == 0 and self.gene5prime.junction > cds[1]:
                            self.effect_5prime = 'intron (before cds)'
                            break
                        elif n == n_max and self.gene5prime.junction < cds[0]:
                            self.effect_5prime = 'intron (after cds)'
                            break


                    if (n!=0) and \
                            (n!=n_max) and \
                            (
                                self.gene5prime.junction < cds[0] or \
                                self.gene5prime.junction > self.transcript1.coding_sequence_position_ranges[n+1][1]
                            ):

                        self.effect_5prime='intron (cds)'
                        break

                    n += 1
            else:
                for cds in self.transcript1.coding_sequence_position_ranges:

                    if self.gene5prime.junction >= cds[0] and self.gene5prime.junction <= cds[1]:
                        self.effect_5prime = 'CDS'
                        break

                if self.transcript1.strand == "+":
                    if self.gene5prime.junction==self.transcript1.coding_sequence_position_ranges[0][0]:
                        self.effect_5prime = 'CDS (start)'
                    elif self.gene5prime.junction==self.transcript1.coding_sequence_position_ranges[-1][1]:
                        self.effect_5prime = 'CDS (end)'
                else:
                    if self.gene5prime.junction==self.transcript1.coding_sequence_position_ranges[0][1]:
                        self.effect_5prime = 'CDS (start)'
                    elif self.gene5prime.junction==self.transcript1.coding_sequence_position_ranges[-1][0]:
                        self.effect_5prime = 'CDS (end)'

        if self.transcript2.complete and (self.effect_3prime.find('UTR') == -1):

            if self.effect_3prime.find('intron') != -1:

                #if in intron, is it between CDS regions?

                n = 0
                n_max = len(self.transcript2.coding_sequence_position_ranges)-1

                for cds in self.transcript2.coding_sequence_position_ranges:

                    if self.transcript2.strand=="+":
                        if n==0 and self.gene3prime.junction < cds[0]:
                            self.effect_3prime='intron (before cds)'
                            break
                        elif n==n_max and self.gene3prime.junction > cds[1]:
                            self.effect_3prime='intron (after cds)'
                            break
                    else:
                        if n == 0 and self.gene3prime.junction > cds[1]:
                            self.effect_3prime = 'intron (before cds)'
                            break
                        elif n == n_max and self.gene3prime.junction < cds[0]:
                            self.effect_3prime = 'intron (after cds)'
                            break

                    if (n != 0) and \
                            (n != n_max) and \
                            (
                                self.gene3prime.junction < cds[0] or \
                                self.gene3prime.junction > self.transcript2.coding_sequence_position_ranges[n+1][1]
                            ):

                        self.effect_3prime = 'intron (cds)'
                        break

                    n += 1

            else:
                for cds in self.transcript2.coding_sequence_position_ranges:
                    if self.gene3prime.junction >= cds[0] and self.gene3prime.junction <= cds[1]:
                        self.effect_3prime='CDS'
                        break

                if self.transcript2.strand == "+":
                    if self.gene3prime.junction==self.transcript2.coding_sequence_position_ranges[0][0]:
                        self.effect_3prime = 'CDS (start)'
                    elif self.gene3prime.junction==self.transcript2.coding_sequence_position_ranges[-1][1]:
                        self.effect_3prime = 'CDS (end)'
                else:
                    if self.gene3prime.junction==self.transcript2.coding_sequence_position_ranges[0][1]:
                        self.effect_3prime = 'CDS (start)'
                    elif self.gene3prime.junction==self.transcript2.coding_sequence_position_ranges[-1][0]:
                        self.effect_3prime = 'CDS (end)'

        # get if has coding potential

        self.has_coding_potential = utils.CODING_COMBINATIONS[(self.effect_5prime,self.effect_3prime)]['protein_coding_potential']

        # check if they don't have stop and/or start codons

        reasons = []

        if not self.transcript1.contains_start_codon:
            self.has_coding_potential=False
            reasons.append("no known 5' transcript start codon")
        if not self.transcript1.contains_stop_codon:
            self.has_coding_potential=False
            reasons.append("no known 5' transcript stop codon")

        if not self.transcript2.contains_start_codon:
            self.has_coding_potential=False
            reasons.append("no known 3' transcript start codon")
        if not self.transcript2.contains_stop_codon:
            self.has_coding_potential=False
            reasons.append("no known 3' transcript stop codon")

        # append information to cdna fasta headers

        self.cdna.description += "; 5' location: " + self.effect_5prime + \
            "; 3' location: " + self.effect_3prime + \
            "; Has protein coding potential: " + str(self.has_coding_potential)

        if not self.has_coding_potential:
            self.cdna.description += "; Reason: " + ', '.join(reasons)

        # if the fusion transcript has coding potential then
        # fetch its CDS and protein sequences and annotate it

        if self.has_coding_potential:
            self._fetch_transcript_cds()
            self._fetch_protein()
            self._annotate()




class DataBaseError(Exception):
    def __init__(self,e):
        Exception.__init__(self,e)
        self.e = e

class GeneIDException(Exception):
    def __init__(self,gene):
        Exception.__init__(
            self,
            "No Ensembl ID found for {}! Check its spelling and if you are " \
            "using the right genome build.".format(gene)
        )

class TooManyGenesException(Exception):
    def __init__(self,gene,ids,build):
        Exception.__init__(
            self,
            "Multiple Ensembl IDs found matching {}: {} for genome {}." \
            " Specify which Ensembl ID.".format(gene,', '.join(ids),build)
        )

class JunctionException(Exception):
    def __init__(self,gene,junction):
        Exception.__init__(
            self,
            "Junction {} not within {} gene boundaries!"
            .format(junction,gene)
        )



# all combinations except those that do not produce some sort of
# chimeric protein are commented out

import itertools
import os

agfusion_db = os.path.join(
    os.path.split(__file__)[0],
    'data',
    'agfusion.db'
)

AGFUSION_DB_URL = "https://github.com/murphycj/AGFusionDB/raw/master/agfusion."

# this is mostly contigent on the maximum ensembl release supported
# by pyensembl

GENOME_SHORTCUTS = {
    'GRCm38':['mus_musculus',92],
    'mm10':['mus_musculus',92],
    'mm9':['mus_musculus',67],
    'GRCh38':['homo_sapiens',92],
    'hg38':['homo_sapiens',92],
    'hg19':['homo_sapiens',75]
}

MAX_ENSEMBL_RELEASE = 92

AVAILABLE_ENSEMBL_SPECIES = {
    'homo_sapiens':range(69,MAX_ENSEMBL_RELEASE+1),
    'mus_musculus':range(67,MAX_ENSEMBL_RELEASE+1)
}

ENSEMBL_MYSQL_TABLES = {
    'homo_sapiens':{},
    'mus_musculus':{}
}

#ENSEMBL_MYSQL_TABLES['homo_sapiens'][48] = 'homo_sapiens_core_48_36j'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][49] = 'homo_sapiens_core_49_36k'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][50] = 'homo_sapiens_core_50_36l'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][51] = 'homo_sapiens_core_51_36m'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][52] = 'homo_sapiens_core_52_36n'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][53] = 'homo_sapiens_core_53_36o'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][54] = 'homo_sapiens_core_54_36p'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][55] = 'homo_sapiens_core_55_37'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][56] = 'homo_sapiens_core_56_37a'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][57] = 'homo_sapiens_core_57_37b'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][58] = 'homo_sapiens_core_58_37c'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][59] = 'homo_sapiens_core_59_37d'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][60] = 'homo_sapiens_core_60_37e'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][61] = 'homo_sapiens_core_61_37f'
#ENSEMBL_MYSQL_TABLES['homo_sapiens'][62] = 'homo_sapiens_core_62_37g'

for i in range(63,MAX_ENSEMBL_RELEASE+1):
    if i < 76:
        ENSEMBL_MYSQL_TABLES['homo_sapiens'][i] = 'homo_sapiens_core_' + str(i) + '_37'
    else:
        ENSEMBL_MYSQL_TABLES['homo_sapiens'][i] = 'homo_sapiens_core_' + str(i) + '_38'

for i in range(67,MAX_ENSEMBL_RELEASE+1):
    if i < 68:
        ENSEMBL_MYSQL_TABLES['mus_musculus'][i] = 'mus_musculus_core_' + str(i) + '_37'
    else:
        ENSEMBL_MYSQL_TABLES['mus_musculus'][i] = 'mus_musculus_core_' + str(i) + '_38'

# min amino acid length of domain to plot it

MIN_DOMAIN_LENGTH = 5

# just

STANDARD_CHROMOSOMES = [str(i) for i in range(1,23)] + ['X','Y','MT']

# the available protein domain annotations

PROTEIN_ANNOTATIONS = [
    'pfam', 'smart', 'superfamily', 'tigrfam', 'pfscan',
    'tmhmm', 'seg', 'ncoils', 'prints',
    'pirsf', 'signalp'
]

JUNCTION_LOCATIONS = [
    'CDS','CDS (start)','CDS (end)','5UTR','5UTR (end)',
    '3UTR','3UTR (start)','exon','intron','intron (cds)',
    'intron (before cds)','intron (after cds)'
    ]

CODING_COMBINATIONS = list(itertools.product(JUNCTION_LOCATIONS,JUNCTION_LOCATIONS))
CODING_COMBINATIONS = {i:{'protein_coding_potential':False,'truncated':False} for i in CODING_COMBINATIONS}


CODING_COMBINATIONS['CDS','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS','CDS (end)'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS','5UTR'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS','5UTR (end)'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS','3UTR'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS','3UTR (start)'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS','intron'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS','intron (before cds)'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS','intron (after cds)'] = {'protein_coding_potential':True,'truncated':True}

CODING_COMBINATIONS['CDS (end)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','5UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','3UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','intron'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (end)','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}

#CODING_COMBINATIONS['CDS (start)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (start)','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['CDS (start)','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['CDS (start)','5UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['CDS (start)','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['CDS (start)','3UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['CDS (start)','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (start)','intron'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS (start)','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['CDS (start)','intron (before cds)'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['CDS (start)','intron (after cds)'] = {'protein_coding_potential':True,'truncated':True}

#CODING_COMBINATIONS['5UTR','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['5UTR','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR','5UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR','3UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR','intron'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['5UTR','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}

#CODING_COMBINATIONS['5UTR (end)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['5UTR (end)','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR (end)','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR (end)','5UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR (end)','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR (end)','3UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR (end)','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR (end)','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['5UTR (end)','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['5UTR (end)','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}

CODING_COMBINATIONS['3UTR','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','5UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','3UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','intron'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}

CODING_COMBINATIONS['3UTR (start)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','5UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','3UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','intron'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (start)','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}

#CODING_COMBINATIONS['3UTR (end)','CDS'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','5UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','3UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','intron'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['3UTR (end)','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['3UTR (end)','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}

#CODING_COMBINATIONS['intron','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron','5UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron','3UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron','intron'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron','intron (after cds)'] = {'protein_coding_potential':{'protein_coding_potential':True,'truncated':False},'truncated':False}

CODING_COMBINATIONS['intron (cds)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (cds)','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (cds)','CDS (end)'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['intron (cds)','5UTR'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['intron (cds)','5UTR (end)'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['intron (cds)','3UTR'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['intron (cds)','3UTR (start)'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['intron (cds)','intron'] = {'protein_coding_potential':True,'truncated':True}
CODING_COMBINATIONS['intron (cds)','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron (cds)','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron (cds)','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}

#CODING_COMBINATIONS['intron (before cds)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (before cds)','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron (before cds)','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (before cds)','5UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (before cds)','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron (before cds)','3UTR'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron (before cds)','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron (before cds)','intron'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron (before cds)','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (before cds)','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
#CODING_COMBINATIONS['intron (before cds)','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}

CODING_COMBINATIONS['intron (after cds)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','CDS (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','5UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','3UTR'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','intron'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','intron (before cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS['intron (after cds)','intron (after cds)'] = {'protein_coding_potential':True,'truncated':False}



"""
Parses output files from fusion-finding algorithms
"""

import re

class _Parser(object):
    def __init__(self, logger):
        self.fusions = []
        self.iterator = 0
        self.logger = logger

    def __iter__(self):
        return self

    def __next__(self):
        if self.iterator >= len(self.fusions):
            raise StopIteration
        else:
            self.iterator += 1
            return self.fusions[self.iterator-1]

    def _check_data(self):
        if len(self.fusions)==0:
            self.logger.error("Read 0 fusions from the file! Exiting...")
        else:
            self.logger.info(
                "Read {} fusions from the file.".format(len(self.fusions))
            )
    next = __next__

class STARFusion(_Parser):
    def __init__(self,infile,logger):
        super(STARFusion, self).__init__(logger)

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('^#',line):
                line = line.rstrip().split('\t')
                if line[0]!='#FusionName' and line[0]!='#fusion_name':
                    raise AssertionError('Unrecognized STAR-Fusion input for first column in header. Should be #FusionName or #fusion_name.')

                assert line[4]=='LeftGene', 'Unrecognized STAR-Fusion input'
                assert line[5]=='LeftBreakpoint', 'Unrecognized STAR-Fusion input'
                assert line[6]=='RightGene', 'Unrecognized STAR-Fusion input'
                assert line[7]=='RightBreakpoint', 'Unrecognized STAR-Fusion input'
                continue

            line = line.strip().split('\t')

            gene_5prime = line[4].split('^')[1].split('.')[0]
            gene_5prime_name = line[4].split('^')[0]
            gene_5prime_junction = int(line[5].split(':')[1])
            gene_3prime = line[6].split('^')[1].split('.')[0]
            gene_3prime_name = line[6].split('^')[0]
            gene_3prime_junction = int(line[7].split(':')[1])
            self.fusions.append(
                {
                    'gene5prime':gene_5prime,
                    'gene3prime':gene_3prime,
                    'alternative_name_5prime':gene_5prime_name,
                    'alternative_name_3prime':gene_3prime_name,
                    'gene5prime_junction':gene_5prime_junction,
                    'gene3prime_junction':gene_3prime_junction
                }
            )
        fin.close()

        self._check_data()

class EricScript(_Parser):
    def __init__(self,infile,logger):
        super(EricScript, self).__init__(logger)

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('^GeneName1',line):
                line = line.strip().split('\t')
                for i, j in zip([3,6,8,9],['Breakpoint1','Breakpoint2','EnsemblGene1','EnsemblGene2']):
                    assert line[i]==j, "Unrecognized EricScript input: {}".format(j)
            else:
                line = line.strip().split('\t')

                gene_5prime_name = line[8]
                gene_5prime_junction = int(line[3])
                gene_3prime_name = line[9]
                gene_3prime_junction = int(line[6])
                self.fusions.append(
                    {
                        'gene5prime':None,
                        'gene3prime':None,
                        'alternative_name_5prime':gene_5prime_name,
                        'alternative_name_3prime':gene_3prime_name,
                        'gene5prime_junction':gene_5prime_junction,
                        'gene3prime_junction':gene_3prime_junction
                    }
                )
        fin.close()

        self._check_data()

class FusionCatcher(_Parser):
    def __init__(self,infile,logger):
        super(FusionCatcher, self).__init__(logger)

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('^Gene_1_symbol',line):
                line = line.rstrip().split('\t')
                assert line[8]=='Fusion_point_for_gene_1(5end_fusion_partner)', 'Unrecognized FusionCatcher input'
                assert line[9]=='Fusion_point_for_gene_2(3end_fusion_partner)', 'Unrecognized FusionCatcher input'
                assert line[10]=='Gene_1_id(5end_fusion_partner)', 'Unrecognized FusionCatcher input'
                assert line[11]=='Gene_2_id(3end_fusion_partner)', 'Unrecognized FusionCatcher input'
                continue

            line = line.strip().split('\t')
            self.fusions.append(
                {
                    'gene5prime':line[10],
                    'gene3prime':line[11],
                    'alternative_name_5prime':line[0],
                    'alternative_name_3prime':line[1],
                    'gene5prime_junction':int(line[8].split(':')[1]),
                    'gene3prime_junction':int(line[9].split(':')[1])
                }
            )
        fin.close()

        self._check_data()

class FusionHunter(_Parser):
    def __init__(self,infile,logger):
        super(FusionHunter, self).__init__(logger)

        gene1 = gene2 = None
        gene1_junction = gene2_junction = None

        fin = open(infile,'r')
        for line in fin.readlines():

            if re.findall('^# Fusion:',line):

                if gene1 is not None and gene2 is not None:
                    self.fusions.append(
                        {
                            'gene5prime':None,
                            'gene3prime':None,
                            'alternative_name_5prime':gene1,
                            'alternative_name_3prime':gene2,
                            'gene5prime_junction':int(gene1_junction),
                            'gene3prime_junction':int(gene2_junction)
                        }
                    )

                strands = re.findall("(?<=\[)(.*)(?=\])",line)
                assert len(strands)==1,"Unrecognized FusionHunter input. Incorrect strand information."
                gene1_strand = strands[0][0]
                gene2_strand = strands[0][1]

            elif re.findall('^--',line):
                #new breakpoint
                if gene1 is not None and gene2 is not None:
                    self.fusions.append(
                        {
                            'gene5prime':None,
                            'gene3prime':None,
                            'alternative_name_5prime':gene1,
                            'alternative_name_3prime':gene2,
                            'gene5prime_junction':int(gene1_junction),
                            'gene3prime_junction':int(gene2_junction)
                        }
                    )

            elif re.findall('^->',line):
                junctions = re.findall("(chr[0-9]*):([0-9]*)-([0-9]*)",line)
                assert len(junctions)==2,"Unrecognized FusionHunter input. Incorrect junction information."

                gene1 = gene2 = None
                gene1_junction = gene2_junction = None

                gene1, gene2 = re.findall("[A-Z0-9]*\sx\s[A-Z0-9a-z]*(?=\t)",line)[0].split(' x ')
                assert gene1 is not None and gene2 is not None, "Unrecognized FusionHunter input. Incorrect gnee information."

                if gene1_strand=='+':
                    gene1_junction = junctions[0][2]
                else:
                    gene1_junction = junctions[0][1]

                if gene2_strand=='+':
                    gene2_junction = junctions[1][1]
                else:
                    gene2_junction = junctions[1][2]
        fin.close()

        if gene1 is not None and gene2 is not None:
            self.fusions.append(
                {
                    'gene5prime':None,
                    'gene3prime':None,
                    'alternative_name_5prime':gene1,
                    'alternative_name_3prime':gene2,
                    'gene5prime_junction':int(gene1_junction),
                    'gene3prime_junction':int(gene2_junction)
                }
            )

        self._check_data()


class FusionMap(_Parser):
    def __init__(self,infile,logger):
        super(FusionMap, self).__init__(logger)

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('^FusionID',line):
                line = line.strip().split('\t')
                for i, j in zip([6,8,9,13],['Position1','Position2','KnownGene1','KnownGene2']):
                    assert line[i]==j, "Unrecognized FusionMap input {} header not in expected position.".format(j)
            else:
                line = line.strip().split('\t')

                gene_5prime_name = line[9]
                gene_5prime_junction = int(line[6])
                gene_3prime_name = line[13]
                gene_3prime_junction = int(line[8])
                self.fusions.append(
                    {
                        'gene5prime':None,
                        'gene3prime':None,
                        'alternative_name_5prime':gene_5prime_name,
                        'alternative_name_3prime':gene_3prime_name,
                        'gene5prime_junction':gene_5prime_junction,
                        'gene3prime_junction':gene_3prime_junction
                    }
                )
        fin.close()

        self._check_data()

class MapSplice(_Parser):
    def __init__(self,infile,logger):
        super(MapSplice, self).__init__(logger)

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('^chrom',line):
                line = line.strip().split('\t')
                for i, j in zip([1,2,60,61],['doner_end','acceptor_start','annotated_gene_donor','annotated_gene_acceptor']):
                    assert line[i]==j, "Unrecognized MapSplice input. {} header not in expected position.".format(j)
            else:
                line = line.strip().split('\t')

                gene_5prime_name = line[60]
                gene_5prime_junction = int(line[1])
                gene_3prime_name = line[61]
                gene_3prime_junction = int(line[2])
                self.fusions.append(
                    {
                        'gene5prime':None,
                        'gene3prime':None,
                        'alternative_name_5prime':gene_5prime_name,
                        'alternative_name_3prime':gene_3prime_name,
                        'gene5prime_junction':gene_5prime_junction,
                        'gene3prime_junction':gene_3prime_junction
                    }
                )
        fin.close()

        self._check_data()

class TopHatFusion(_Parser):
    def __init__(self,infile,logger):
        super(TopHatFusion, self).__init__(logger)

        fin = open(infile,'r')
        for line in fin.readlines():
            line = line.strip().split('\t')

            gene_5prime = line[1]
            gene_5prime_name = line[1]
            gene_5prime_junction = int(line[3])
            gene_3prime = line[4]
            gene_3prime_name = line[4]
            gene_3prime_junction = int(line[6])
            self.fusions.append(
                {
                    'gene5prime':gene_5prime,
                    'gene3prime':gene_3prime,
                    'alternative_name_5prime':gene_5prime_name,
                    'alternative_name_3prime':gene_3prime_name,
                    'gene5prime_junction':gene_5prime_junction,
                    'gene3prime_junction':gene_3prime_junction
                }
            )
        fin.close()

        self._check_data()

class DeFuse(_Parser):
    def __init__(self,infile,logger):
        super(DeFuse, self).__init__(logger)

        data_indices = {
            'gene5prime':None,
            'gene3prime':None,
            'gene5prime_junction':None,
            'gene3prime_junction':None
        }

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('^cluster_id',line):
                line = line.strip().split('\t')
                for column in data_indices.keys():
                    try:
                        data_indices[column] = line.index(column)
                    except ValueError:
                        logger.error(
                            "Unrecognized {} input! Cannot find {} column."
                            .format(
                                self.__class__.__name__,
                                column
                            )
                        )
                        exit()
                continue
            if not any([i is None for i in data_indices.keys()]):
                line = line.strip().split('\t')
                self.fusions.append(
                    {
                        'gene5prime':line[data_indices['gene5prime']],
                        'gene3prime':line[data_indices['gene3prime']],
                        'gene5prime_junction':int(line[data_indices['gene5prime_junction']]),
                        'gene3prime_junction':int(line[data_indices['gene3prime_junction']])
                    }
                )
        fin.close()

        self._check_data()

class Chimerascan(_Parser):
    def __init__(self,infile,logger):
        super(Chimerascan, self).__init__(logger)

        data_indices = {
            'genes5p':12,
            'genes3p':13,
            'start5p':1,
            'end5p':2,
            'start3p':4,
            'end3p':5,
            'strand5p':8,
            'strand3p':9
        }

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('#chrom5p',line):
                line = line.strip().split('\t')
                for column in data_indices.keys():
                    try:
                        data_indices[column] = line.index(column)
                    except ValueError:
                        logger.error(
                            "Unrecognized {} input! Cannot find {} column."
                            .format(
                                self.__class__.__name__,
                                column
                            )
                        )
                        exit()
                continue

            line = line.strip().split('\t')

            if line[data_indices['strand5p']]=='+':
                gene1_junction = line[data_indices['end5p']]
            else:
                gene1_junction = line[data_indices['start5p']]

            if line[data_indices['strand3p']]=='+':
                gene2_junction = line[data_indices['start3p']]
            else:
                gene2_junction = line[data_indices['end3p']]

            self.fusions.append(
                {
                    'gene5prime':line[data_indices['genes5p']].split(','),
                    'gene3prime':line[data_indices['genes3p']].split(','),
                    'gene5prime_junction':int(gene1_junction),
                    'gene3prime_junction':int(gene2_junction)
                }
            )
        fin.close()

        self._check_data()

class ChimeRScope(_Parser):
    def __init__(self,infile,logger):
        super(ChimeRScope, self).__init__(logger)

        data_indices = {
            'Gene1':2,
            'Gene2':4,
            'Gene1_fusionPoint':7,
            'Gene2_fusionPoint':9
        }

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('^ConfidentScore',line):
                line = line.strip().split('\t')
                for column in data_indices.keys():
                    try:
                        data_indices[column] = line.index(column)
                    except ValueError:
                        logger.error(
                            "Unrecognized {} input! Cannot find {} column."
                            .format(
                                self.__class__.__name__,
                                column
                            )
                        )
                        exit()
                continue

            line = line.strip().split('\t')
            self.fusions.append(
                {
                    'gene5prime':line[data_indices['Gene1']],
                    'gene3prime':line[data_indices['Gene2']],
                    'gene5prime_junction':line[data_indices['Gene1_fusionPoint']],
                    'gene3prime_junction':line[data_indices['Gene2_fusionPoint']]
                }
            )
        fin.close()

        self._check_data()

class JAFFA(_Parser):
    def __init__(self,infile,logger):
        super(JAFFA, self).__init__(logger)

        data_indices = {
            'base1':7,
            'base2':9,
            'fusion genes':1
        }

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('sample',line):
                line = line.strip().replace('"','').split(',')
                for column in data_indices.keys():
                    try:
                        data_indices[column] = line.index(column)
                    except ValueError:
                        logger.error(
                            "Unrecognized {} input! Cannot find {} column."
                            .format(
                                self.__class__.__name__,
                                column
                            )
                        )
                        exit()
                continue

            line = line.strip().replace('"','').split(',')
            self.fusions.append(
                {
                    'gene5prime':line[data_indices['fusion genes']].split(':')[0],
                    'gene3prime':line[data_indices['fusion genes']].split(':')[1],
                    'gene5prime_junction':int(line[data_indices['base1']]),
                    'gene3prime_junction':int(line[data_indices['base2']])
                }
            )
        fin.close()

        self._check_data()

class Bellerophontes(_Parser):
    def __init__(self,infile,logger):
        super(Bellerophontes, self).__init__(logger)

        fin = open(infile,'r')
        for line in fin.readlines():

            line = line.strip().split('\t')
            if len(line) <= 2:
                continue

            gene1_strand = line[2]
            gene2_strand = line[6]

            self.fusions.append(
                {
                    'gene5prime':line[0],
                    'gene3prime':line[4],
                    'gene5prime_junction':int(line[9]),
                    'gene3prime_junction':int(line[11])
                }
            )
        fin.close()

        self._check_data()

class BreakFusion(_Parser):
    def __init__(self,infile,logger):
        super(BreakFusion, self).__init__(logger)

        data_indices = {
            'POS1':1,
            'POS2':4,
            'RefseqGene':11
        }

        fin = open(infile,'r')
        for line in fin.readlines():
            if re.findall('CHR1',line):
                line = line.strip().split('\t')
                for column in data_indices.keys():
                    try:
                        data_indices[column] = line.index(column)
                    except ValueError:
                        logger.error(
                            "Unrecognized {} input! Cannot find {} column."
                            .format(
                                self.__class__.__name__,
                                column
                            )
                        )
                        exit()
                continue

            if re.findall('Fusion',line):
                line = line.strip().split('\t')
                genes = line[data_indices['RefseqGene']]
                genes = re.findall("(?<=Gene:)(.*?)(?=,)",genes)
                if len(genes)!=1:
                    self.logger.error('Could not parse genes from BreakFusion input line: {}'.format('\t'.join(line)))
                genes = genes[0].split('|')
                if len(genes)!=2:
                    self.logger.error('Could not find two genes for BreakFusion input line: {}'.format('\t'.join(line)))

                self.fusions.append(
                    {
                        'gene5prime':genes[0],
                        'gene3prime':genes[1],
                        'gene5prime_junction':int(line[data_indices['POS1']]),
                        'gene3prime_junction':int(line[data_indices['POS2']])
                    }
                )
        fin.close()

        self._check_data()

class InFusion(_Parser):
    def __init__(self,infile,logger):
        super(InFusion, self).__init__(logger)

        data_indices = {
            'break_pos1':2,
            'break_pos2':5,
            'genes_1':9,
            'genes_2':10
        }

        fin = open(infile,'r')
        n=0
        for line in fin.readlines():
            n+=1
            if re.findall('#id',line):
                line = line.strip().split('\t')
                for column in data_indices.keys():
                    try:
                        data_indices[column] = line.index(column)
                    except ValueError:
                        logger.error(
                            "Unrecognized {} input! Cannot find {} column."
                            .format(
                                self.__class__.__name__,
                                column
                            )
                        )
                        exit()
                continue

            line = line.strip().split('\t')

            if line[data_indices['genes_1']]=='none' or \
               line[data_indices['genes_2']]=='none':
                self.logger.warn(
                    'Skipping fusion on line {} because one or more ' \
                    'of the provided gene names under \'gene_1\' and' \
                    ' \'gene_2\' is listed as \'none\'.'
                    .format(n)
                )
                continue

            self.fusions.append(
                {
                    'gene5prime':line[data_indices['genes_1']].split(';'),
                    'gene3prime':line[data_indices['genes_2']].split(';'),
                    'gene5prime_junction':int(line[data_indices['break_pos1']]),
                    'gene3prime_junction':int(line[data_indices['break_pos2']])
                }
            )
        fin.close()

        self._check_data()

class NFuse(_Parser):
    def __init__(self,infile,logger):
        super(NFuse, self).__init__(logger)

        self._check_data()

class SOAPfuse(_Parser):
    def __init__(self,infile,logger):
        super(SOAPfuse, self).__init__(logger)

        self._check_data()

class FusionEntry():
    def __init__(self,infile,logger):
        super(FusionEntry, self).__init__(logger)

        self._check_data()

parsers = {
    'bellerophontes':Bellerophontes,
    'breakfusion':BreakFusion,
    'chimerascan':Chimerascan,
    'chimerscope':ChimeRScope,
    'defuse':DeFuse,
    'ericscript':EricScript,
    'fusioncatcher':FusionCatcher,
    'fusionhunter':FusionHunter,
    'fusionmap':FusionMap,
    'infusion':InFusion,
    'jaffa':JAFFA,
    #'fusionseq':FusionSeq,
    #'prada':Prada,
    #'gfusion':GFusion,
    #'machete':Machete,
    'mapsplice':MapSplice,
    #'nfuse':nFuse,
    #'soapfuse':SOAPfuse,
    'starfusion':STARFusion,
    'tophatfusion':TopHatFusion
}
